__version__ = '3.0.0'

from .resource import JSONResource
from .conf import APIConfig, ResourceConfig, BodyParameter, QueryParameter
from .exception import  RestClientConfigurationError
from .resource import API

